@@ -1,4 +1,3 @@
-from dataclasses import dataclass
 from io import BytesIO
 from pprint import pformat
 from unittest.mock import call, patch, mock_open
@@ -8,7 +7,8 @@ from pytest import fixture, raises
 from pymw import API, LoginError, APIError
 
 
-api = API('https://www.mediawiki.org/w/api.php')
+url = 'https://www.mediawiki.org/w/api.php'
+api = API(url)
 
 
 @fixture
@@ -21,10 +21,12 @@ def fake_sleep(_):
     return
 
 
-@dataclass
 class FakeResp:
-    headers: dict
-    _json: dict
+    __slots__ = ('_json', 'headers')
+
+    def __init__(self, json, headers=None):
+        self._json = json
+        self.headers = {} if headers is None else headers
 
     def json(self):
         return self._json
@@ -49,11 +51,21 @@ def api_post_patch(*call_returns):
 
 
 def session_post_patch(*call_header_returns):
-    chr = iter(call_header_returns)
-    call_returns = []
-    for call, headers, return_ in zip(chr, chr, chr):
-        call_returns += (call, FakeResp(headers, return_))
-    return patch_post(api.session, 'post', call_returns)
+    call_responses = []
+    iterator = iter(call_header_returns)
+    call = next(iterator)
+    while call is not None:
+        headers_or_json = next(iterator)
+        json_or_call = next(iterator, None)
+        if type(json_or_call) is dict:
+            response = FakeResp(json=json_or_call, headers=headers_or_json)
+            call_responses += (call, response)
+            call = next(iterator, None)
+        else:
+            response = FakeResp(json=headers_or_json)
+            call_responses += (call, response)
+            call = json_or_call
+    return patch_post(api.session, 'request', call_responses)
 
 
 @api_post_patch(
@@ -105,9 +117,15 @@ def test_recentchanges(_):
 @patch('pymw._api.sleep', fake_sleep)
 @patch('pymw._api.warning')
 @session_post_patch(
-    call(api.url, data={'action': 'query', 'errorformat': 'plaintext', 'format': 'json', 'formatversion': '2', 'maxlag': 5, 'meta': 'tokens', 'type': 'watch'}, files=None),
+    call('POST', url, data={
+        'action': 'query', 'errorformat': 'plaintext', 'format': 'json',
+        'formatversion': '2', 'maxlag': 5, 'meta': 'tokens', 'type': 'watch'
+    }, files=None),
     {'retry-after': '5'}, {'errors': [{'code': 'maxlag', 'text': 'Waiting for 10.64.16.7: 0.80593395233154 seconds lagged.', 'data': {'host': '10.64.16.7', 'lag': 0.805933952331543, 'type': 'db'}, 'module': 'main'}], 'docref': 'See https://www.mediawiki.org/w/api.php for API usage. Subscribe to the mediawiki-api-announce mailing list at &lt;https://lists.wikimedia.org/mailman/listinfo/mediawiki-api-announce&gt; for notice of API deprecations and breaking changes.', 'servedby': 'mw1225'},
-    call(api.url, data={'meta': 'tokens', 'type': 'watch', 'action': 'query', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}, files=None),
+    call('POST', url, data={
+        'meta': 'tokens', 'type': 'watch', 'action': 'query', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5
+    }, files=None),
     {}, {'batchcomplete': True, 'query': {'tokens': {'watchtoken': '+\\'}}})
 def test_maxlag(_, warning_mock, cleared_api):
     tokens = cleared_api.query_meta('tokens', type='watch')
@@ -125,10 +143,19 @@ def test_siteinfo(post_mock):
 
 
 @api_post_patch(
-    call({'action': 'query', 'prop': 'langlinks', 'lllimit': 1, 'titles': 'Main Page'}),
-    {'continue': {'llcontinue': '15580374|bg', 'continue': '||'}, 'query': {'pages': [{'pageid': 15580374, 'ns': 0, 'title': 'Main Page', 'langlinks': [{'lang': 'ar', 'title': ''}]}]}},
-    call({'action': 'query', 'prop': 'langlinks', 'lllimit': 1, 'titles': 'Main Page', 'llcontinue': '15580374|bg', 'continue': '||'}),
-    {'batchcomplete': True, 'query': {'pages': [{'pageid': 15580374, 'ns': 0, 'title': 'Main Page', 'langlinks': [{'lang': 'zh', 'title': ''}]}]}})
+    call({
+        'action': 'query', 'prop': 'langlinks', 'lllimit': 1,
+        'titles': 'Main Page'}),
+    {'continue': {'llcontinue': '15580374|bg', 'continue': '||'}, 'query': {
+        'pages': [{
+            'pageid': 15580374, 'ns': 0, 'title': 'Main Page',
+            'langlinks': [{'lang': 'ar', 'title': ''}]}]}},
+    call({
+        'action': 'query', 'prop': 'langlinks', 'lllimit': 1,
+        'titles': 'Main Page', 'llcontinue': '15580374|bg', 'continue': '||'}),
+    {'batchcomplete': True, 'query': {'pages': [{
+        'pageid': 15580374, 'ns': 0, 'title': 'Main Page',
+        'langlinks': [{'lang': 'zh', 'title': ''}]}]}})
 def test_langlinks(_):
     titles_langlinks = [page_ll for page_ll in api.query_prop(
         'langlinks', titles='Main Page', lllimit=1)]
@@ -174,10 +201,10 @@ def test_context_manager():
 
 @session_post_patch(
     any, {}, {'batchcomplete': True, 'query': {'tokens': {'patroltoken': '+\\'}}},
-    call(
-        api.url,
-        data={'revid': 27040231, 'action': 'patrol', 'token': '+\\', 'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5},
-        files=None), {}, {'errors': [{'code': 'permissiondenied', 'text': 'T', 'module': 'patrol'}], 'docref': 'D', 'servedby': 'mw1233'})
+    call('POST', url, data={
+        'revid': 27040231, 'action': 'patrol', 'token': '+\\', 'format':
+            'json', 'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5
+    }, files=None), {}, {'errors': [{'code': 'permissiondenied', 'text': 'T', 'module': 'patrol'}], 'docref': 'D', 'servedby': 'mw1233'})
 def test_patrol_not_logged_in(_, cleared_api):
     try:
         cleared_api.patrol(revid=27040231)
@@ -351,3 +378,151 @@ def test_assert_login(post_mock):
     api._assert_user = 'USER'
     api.post({})
     assert post_mock.mock_calls[0].kwargs['data']['assertuser'] == 'USER'
+
+
+@patch('pymw._api.Path.open', pymw_toml_mock)
+@session_post_patch(
+    call('POST', url, data={
+        'notfilter': '!read', 'meta': 'notifications', 'action': 'query',
+        'errorformat': 'plaintext', 'format': 'json', 'formatversion': '2',
+        'maxlag': 5}, files=None),
+    {'errors': [{'code': 'login-required', 'text': 'You must be logged in.', 'module': 'query+notifications'}], 'docref': '', 'servedby': 'mw1341'},
+    call('POST', url, data={
+        'type': 'login', 'meta': 'tokens', 'action': 'query', 'format': 'json',
+        'formatversion': '2', 'errorformat': 'plaintext', 'maxlag': 5}, files=None),
+    {'batchcomplete': True, 'query': {'tokens': {'logintoken': 'T1'}}},
+    call('POST', url, data={
+        'action': 'login', 'lgname': 'username@toolname', 'lgpassword':
+            'bot_password', 'lgtoken': 'T1', 'format': 'json', 'formatversion':
+            '2', 'errorformat': 'plaintext', 'maxlag': 5}, files=None),
+    {'login': {'result': 'Success', 'lguserid': 1, 'lgusername': 'username'}},
+    call('POST', url, data={
+        'notfilter': '!read', 'meta': 'notifications', 'action': 'query',
+        'format': 'json', 'formatversion': '2', 'errorformat': 'plaintext',
+        'maxlag': 5, 'assertuser': 'username'}, files=None),
+    {'batchcomplete': True, 'query': {'notifications': {'list': [], 'continue': None}}},
+)
+def test_handle_login_required(_, cleared_api):
+    r = api.query_meta('notifications', notfilter='!read')
+    assert r == {'list': [], 'continue': None}
+
+
+@api_post_patch(
+    any, {'batchcomplete': True, 'continue': {
+        'gapcontinue': '!!', 'continue': 'gapcontinue||'}},
+    any, {'batchcomplete': True, 'continue': {
+        'gapcontinue': '!!!', 'continue': 'gapcontinue||'}, 'query': {
+        'pages': [{'pageid': 3632887, 'ns': 0, 'title': '!!', 'revisions': [{
+            'slots': {'main': {
+                'contentmodel': 'wikitext', 'contentformat': 'text/x-wiki',
+                'content': ...}}}]}]}})
+def test_empty_pages_in_prop_query(_):
+    # should not raise KeyError: 'query'
+    next(api.query_prop('revisions'))
+
+
+page = {
+    'ns': 0,
+    'pageid': 8988,
+    'revisions': [{'slots': {'main': {
+        'content': '', 'contentformat': 'text/x-wiki',
+        'contentmodel': 'wikitext'}}}],
+    'title': 'E (عدد)'}
+
+
+@api_post_patch(
+    any, {'continue': {
+        'continue': 'gapcontinue||', 'gapcontinue': 'Antinatalism',
+        'rvcontinue': '1844356|28824240'}, 'limits': {'allpages': 5000},
+        'query': {'pages': [page]}},
+    any, {'batchcomplete': True, 'continue': {
+        'continue': 'gapcontinue||', 'gapcontinue': 'ISO3166-2:AD'},
+        'limits': {'allpages': 5000}, 'query': {'pages': [{
+            'ns': 0, 'pageid': 8988, 'title': 'E (عدد)'}]}})
+def test_prop_complete_first(_):
+    # used to give KeyError: 'revisions'
+    assert next(api.query_prop('revisions')) is page
+
+
+@api_post_patch(
+    any, {
+        'continue': {
+            'continue': 'gapcontinue||', 'gapcontinue': 'ISO3166-2:AD'},
+        'limits': {'allpages': 5000}, 'query': {
+            'pages': [{
+                'ns': 0, 'pageid': 8988, 'title': 'E (عدد)'}]}},
+    any, {
+        'batchcomplete': True,
+        'continue': {
+            'continue': 'gapcontinue||', 'gapcontinue': 'Antinatalism',
+            'rvcontinue': '1844356|28824240'},
+        'limits': {'allpages': 5000},
+        'query': {'pages': [page]}})
+def test_prop_incomplete_first(_):
+    # used to give KeyError: 'revisions'
+    assert next(api.query_prop('revisions')) == page
+
+
+c2 = {'continue': 'gapcontinue||', 'gapcontinue': 'ISO3166-2:AD'}
+c1 = {
+    'continue': 'gapcontinue||', 'gapcontinue': 'Antinatalism',
+    'rvcontinue': '1844356|28824240'}
+
+
+@api_post_patch(
+    call({}), {'continue': c1}, call(c1), {'continue': c2}, call(c2), {})
+def test_remove_old_continues_from_data(_):
+    for _ in api.post_and_continue({}):
+        pass
+
+
+incomplete_page2 = {'ns': 0, 'pageid': 2393444, 'title': 'آ س رایو ناگار'}
+page2 = {
+    'ns': 0, 'pageid': 2393444, 'revisions': [{'slots': {'main': {
+        'content': '', 'contentformat': 'text/x-wiki',
+        'contentmodel': 'wikitext'}}}],
+    'title': 'آ س رایو ناگار'}
+
+
+@api_post_patch(
+    any, {
+        'continue': {
+            'continue': 'gapcontinue||',
+            'gapcontinue': 'Q_(ابهام\u200cزدایی)',
+            'rvcontinue': '2393444|28481387'},
+        'limits': {'allpages': 5000}, 'query': {'pages': [incomplete_page2]}},
+    any, {
+        'continue': {
+            'continue': 'gapcontinue||',
+            'gapcontinue': 'Q_(ابهام\u200cزدایی)',
+            'rvcontinue': '2393444|28481387'},
+        'limits': {'allpages': 5000},
+        'query': {'pages': [page2]}},
+    any, {
+        'batchcomplete': True, 'continue': {
+            'continue': 'gapcontinue||', 'gapcontinue': 'آئودی_لو_مان_کواترو'},
+        'limits': {'allpages': 5000}, 'query': {'pages': [incomplete_page2]}})
+def test_complete_in_the_middle_of_batch(_):
+    assert next(api.query_prop('revisions')) == page2
+
+
+page3 = {'ns': 0, 'pageid': 1844356, 'revisions': [{
+    'slots': {'main': {
+        'content': '', 'contentformat': 'text/x-wiki',
+        'contentmodel': 'wikitext'}}}], 'title': 'Getopt'}
+
+
+@api_post_patch(
+    any, {
+        'continue': {
+            'continue': 'gapcontinue||', 'gapcontinue': 'Antinatalism',
+            'rvcontinue': '1844356|28824240'},
+        'query': {'pages': [{
+            'ns': 0, 'pageid': 1844356, 'title': 'Getopt'}, page2]}},
+    any, {
+        'batchcomplete': True, 'continue': {
+            'continue': 'gapcontinue||', 'gapcontinue': 'ISO3166-2:AD'},
+        'query': {'pages': [page3, incomplete_page2]}})
+def test_page_in_last_batch(_):
+    page = next(api.query_prop('revisions'))
+    assert page is page3
